<?php

namespace Drupal\gamification\Plugin;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Database\Connection;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\userpoints\Service\UserPointsService;
use Psr\Log\LoggerInterface;

/**
 * Base for a gamification widget that makes use of userpoints.
 */
abstract class GamificationWidgetUserpointsBase extends GamificationWidgetBase {

  /**
   * The logger service.
   *
   * @var \Psr\Log\LoggerInterface $logger
   */
  protected $logger;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $userPointsService;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, ModuleHandlerInterface $module_handler, AccountInterface $current_user, EntityTypeManagerInterface $entity_type_manager, Connection $database, LoggerInterface $logger, UserPointsService $userpoints_service) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $module_handler, $current_user, $entity_type_manager, $database, $userpoints_service);
    $this->logger = $logger;
    $this->userPointsService = $userpoints_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('module_handler'),
      $container->get('current_user'),
      $container->get('entity_type.manager'),
      $container->get('database'),
      $container->get('logger.factory')->get('gamification'),
      $container->get('userpoints.points')
    );
  }

  /**
   * Add points to user and create gamification log.
   *
   * @param \Drupal\gamification\GamificationInterface $gamification_config_entity
   *   Gamification configuration entity.
   * @param string $method
   *   Triggering event string.
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   Content entity which triggered this event.
   * @param int $points
   *   Points that must be attributed to user.
   *
   * @return bool
   *   Flag indicating if operation is executed with success.
   */
  public function addPoints($gamification_config_entity, $method, $entity, $points = 0) {
    $user_points = $points;
    if ($method !== 'entity_delete' && !$this->checkLimits($gamification_config_entity, $method, $entity)) {
      return FALSE;
    }

    $options = $gamification_config_entity->getWidgetOptionsValue();
    if (empty($options)) {
      return FALSE;
    }

    $uid = $this->currentUser->id();
    if ($method === 'entity_delete') {
      $result = $this->getGamificationLogByValues([
        'entity_id' => intval($entity->id()),
      ], ['uid', 'points', 'created']);
      if (empty($result) || !isset($result->uid) || !isset($result->points)) {
        return FALSE;
      }
      $uid = $result->uid;
      $user_points = intval($result->points);
      $user_points = $user_points * (-1);
    }

    $uid = intval($uid);
    if (empty($uid)) {
      return FALSE;
    }

    $user_entity = \Drupal::entityTypeManager()->getStorage('user')->load($uid);
    $log_message = isset($options['log_message']) ? $options['log_message'] : '';
    $userpoints_type = isset($options['userpoints_type']) && !empty($options['userpoints_type']) ? $options['userpoints_type'] : '';

    if (empty($log_message) || empty($userpoints_type)) {
      return FALSE;
    }

    try {
      $this->userPointsService->addPoints($user_points, $userpoints_type, $user_entity, $log_message);
    }
    catch (\Exception $exception) {
      $this->logger->error('Error attributing points to user. @message', ['@message' => $exception->getMessage()]);
    }

    $this->createGamificationLog($gamification_config_entity, $method, $entity, $user_points);

    return TRUE;
  }

}
